#include "Clock.h"
#include "Touch.h"
#include "RTC.h"
#include "fastTFT.h"
#include "SimpleDLS.h"

// the main clock logic


// touch 2 TL & BR diagonal corner cells to config
#define CLOCK_CFG_MASK (word)(0b0001000000001000)

// touch 2 TR & BL diagonal corner cells for alternate (eg mode)
#define CLOCK_ALT_MASK (word)(0b1000000000000001)

// touch any LHS cell to sel (increment)
#define CLOCK_SEL_MASK    (word)(0b0011001100110011)

// touch any RHS cell to set (advance)
#define CLOCK_SET_MASK    (word)(0b1100110011001100)

#define CLOCK_MODE_INDEX (RTC_RAM_BASE_INDEX)

Clock::Clock(TriangleMesh& mesh):
  m_Mesh(mesh),
  m_bEggShown(false)
{
}

void Clock::Init()
{
#ifdef CLOCK_DEBUG  
  Splash(1000, true);
#else
  Splash(5000, true);
  Randomize();
#endif

  m_UpdateTimer = m_HoldoffTimer = millis();
  m_DisplayedMinute = -1;
  m_CfgTouchedCells = 0;
  m_AltTouchedCells = 0;
  m_Mesh.Clear();
  m_DisplayMode = rtc.ReadByte(CLOCK_MODE_INDEX);
  if (m_DisplayMode >= eModeLAST)
  {
    m_DisplayMode = eMode24Hour;
  }
  // Use NZ daylight savings
  SimpleDLS::SetData(SimpleDLS::m_pNewZealandData);
}

void Clock::Loop()
{
  word Touch = touch.TouchedCell();
  
  if (Touch)
  {
    if (Touch & CLOCK_CFG_MASK)
    {
      m_CfgTouchedCells |= Touch;
      if (m_CfgTouchedCells == CLOCK_CFG_MASK)
      {
        m_CfgTouchedCells = 0;
        Config();
        m_DisplayedMinute = -1;
        m_UpdateTimer = millis();
      }
      else
      {
        // feedback about touch
        m_Mesh.RandomizeColour(false);
        m_Mesh.Paint();
        // and hold-off updates
        m_HoldoffTimer = millis();
      }
    }
    else
    {
      // touch elsewhere, clear the mask
      m_CfgTouchedCells = 0;
    }

    if (Touch & CLOCK_ALT_MASK)
    {
      m_AltTouchedCells |= Touch;
      if (m_AltTouchedCells == CLOCK_ALT_MASK)
      {
        m_AltTouchedCells = 0;
        Mode();
        m_DisplayedMinute = -1;
        m_UpdateTimer = millis();
      }
      else
      {
        // feedback about touch
        m_Mesh.RandomizeColour(false);
        m_Mesh.Paint();
        // and hold-off updates
        m_HoldoffTimer = millis();
      }
    }
    else
    {
      // touch elsewhere, clear the mask
      m_AltTouchedCells = 0;
    }
    
    if (m_CfgTouchedCells == 0 && m_AltTouchedCells == 0)
    {
      Randomize();
      // invalid touch
      m_Mesh.RandomizeColour(true);
      m_Mesh.Paint();
      // and hold-off updates
      m_HoldoffTimer = millis();
    }
  }
  
  if (m_CfgTouchedCells || m_AltTouchedCells)
  {
    if (CheckPeriod(m_HoldoffTimer, 15000))
    {
      // hold-off expired, clear touches
      m_AltTouchedCells = m_CfgTouchedCells = 0;
    }
    else
    {
      // hold-off updates
      return;
    }
  }
  
  int Minute = rtc.ReadMinute();
  if (CheckPeriod(m_UpdateTimer, 5000) || m_DisplayedMinute != Minute)
  {
    if (m_DisplayedMinute != Minute)
    {
      // time to update the display
      m_Mesh.Clear();
      rtc.ReadTime();
      rtc.ReadDate();
      int DisplayHour24 = SimpleDLS::GetDisplayHour24(rtc.m_Hour24, rtc.m_DayOfMonth, rtc.m_Month, rtc.m_Year);

#ifdef CLOCK_DEBUG
      if (rtc.m_Minute == 8)  // testing
#else
      if (DisplayHour24 == 12 && rtc.m_Minute == 18)  // easter egg for "December 18th", John's birthday
#endif     
      {
        if (!m_bEggShown)
        {
          Splash(1000, false);
          m_bEggShown = true;
        }
      }
      else
      {
        m_bEggShown = false;
      }
      
      char buffer[10];
      buffer[0] = 0;
      m_Mesh.RandomizeMesh(true);
      m_UpdateTimer = millis();
      if (m_DisplayMode == eMode12Hour)
      {
        Display12Hour(DisplayHour24, buffer);
      }
      else if (m_DisplayMode == eModeTemperature)
      {
        DisplayTemperature(buffer);
      }
      else
      {
        AppendNum(buffer, DisplayHour24, '0');
        AppendNum(buffer, rtc.m_Minute, '0');
        m_Mesh.Display(buffer, 8, 0);
      }
      m_DisplayedMinute = Minute;
    }
    
    m_Mesh.RandomizeColour(true);
    m_Mesh.Paint();
  }  
}

void Clock::Randomize()
{
  rtc.ReadTime();
  rtc.ReadDate();
  long seed = word(rtc.m_Hour24+1, rtc.m_DayOfMonth) | word(rtc.m_Minute+1, rtc. m_Month) << 16;
  randomSeed(seed);
}

size_t Clock::AppendNum(char* Buffer, int Num, char Lead)
{
  // just 0..99
  size_t Len = strlen(Buffer);
  Num = Num % 100;
  if (Num < 10)
  {
    if (Lead)
      Buffer[Len++] = Lead;
  }
  else
  {
    Buffer[Len++] = '0' + (Num / 10);
  }
  Buffer[Len++] = '0' + (Num % 10);
  Buffer[Len] = 0;
  return Len;
}

bool Clock::CheckPeriod(unsigned long& Timer, unsigned long PeriodMS)
{
  // has the timer interval gone off?
  unsigned long Now = millis();
  if (Now - Timer > PeriodMS)
  {
    Timer = Now;
    return true;
  }
  return false;
}

void Clock::Display12Hour(int DisplayHour24, char* buffer)
{
  byte Hr = DisplayHour24;
  if (Hr == 0)
    Hr = 12;
  else if (Hr > 12)
    Hr -= 12;
   
  bool ThinOnes = false; 
  if (Hr >= 10)
  {
    // use a thin '1' to make things fit
    ThinOnes = true;
    m_Mesh.Display("/", 8, 0);
  }
  int col = 2;
  if (Hr == 11)
  {
    strcpy(buffer, "/");  // double-thin
    col++;
  }
  else
  {
    AppendNum(buffer, Hr % 10, '\0'); // no padding char
  }
  m_Mesh.Display(buffer, 8, col);

  // slip in the : between chars
  col = m_Mesh.Display(":", 8, 6);
  col -= 2;

  buffer[0] = 0;
  AppendNum(buffer, rtc.m_Minute, '0');
  if (ThinOnes)
  {
    // convert 1's to san-serif versions
    char* pCh = buffer;
    while (*pCh)
    {
      if (*pCh == '1')
        *pCh = '<';
      pCh++;
    }
  }
  m_Mesh.Display(buffer, 8, col);
}

void Clock::DisplayTemperature(char* buffer)
{
  // not sure how useful this is, the LCD runs warm
  buffer[0] = 0;
  AppendNum(buffer, rtc.ReadTemperature(), ' ');
  strcat(buffer, ";");  // degrees
  m_Mesh.Display(buffer, 8, 2);
}

//                                        D(ay)          M-M(onth)           Y(ear)           h(our)         m-m(inute)
prog_uint32_t Clock::pLabels[] = {0x00000,0xE999E, 0x11111,0xF5555,  0x00000,0x99F1F, 0x00000,0x88F99, 0x00111,0x00F55};

prog_uint32_t Clock::pTime[]   = {0xE4444, 0xBAAAA, 0xEAAAA, 0xE8E8E}; // "TIME"
prog_uint32_t Clock::pDate[]   = {0xCAAAC, 0xEAEAA, 0xE4444, 0xE8E8E}; // "DATE"
prog_uint32_t Clock::pMEW[]    = {0x8DA88, 0xBABAB, 0xA2A3A, 0x22A62}; // "MEW"
prog_uint32_t Clock::pEgg[]    = {0xE444C, 0x00EAE, 0x88EAA, 0x00CAA}; // "John"

void Clock::Splash(int DelayMS, bool Flat)
{
  m_Mesh.Clear();
  if (Flat)
  {
    m_Mesh.RandomizeMesh(false);
    m_Mesh.RandomizeColour(true);
  }
  m_Mesh.Display(pEgg, 4, 8, 1);
  m_Mesh.Paint();
  delay(DelayMS);
  m_Mesh.Clear();
}

void Clock::Config(byte& value, byte Min, byte Max, char Pad, byte LabelIdx)
{
  char buffer[10];
  word Touch;
  bool Display = true;
  do
  {
    if (Display)
    {
      buffer[0] = 0;
      AppendNum(buffer, value, Pad);
      m_Mesh.Clear();
      m_Mesh.Display(buffer, 8, 0);
      m_Mesh.Display(pLabels + LabelIdx*2, 2, 8, 9);
      m_Mesh.Paint();
      Display = false;
    }
    Touch = touch.TouchedCell();
    if (Touch & CLOCK_SEL_MASK)
    {
      value += 1;
      if (value > Max)
        value = Min;
      Display = true;
    }
  } while (!(Touch & CLOCK_SET_MASK));
}

void Clock::Config()
{
  rtc.ReadTime();
  rtc.ReadDate();
  m_Mesh.RandomizeMesh(false);
  m_Mesh.RandomizeColour(false);
  // date
  m_Mesh.Clear();
  m_Mesh.Display(pDate, 4, 8, 0);
  m_Mesh.Paint();
  delay(1000);
  Config(rtc.m_DayOfMonth, 1, 31, ' ', 0);
  Config(rtc.m_Month, 1, 12, ' ', 1);
  Config(rtc.m_Year, 14, 24, ' ', 2);
  rtc.WriteDate();
  
  // time
  m_Mesh.Clear();
  m_Mesh.Display(pTime, 4, 8, 0);
  m_Mesh.Paint();
  delay(1000);
  byte DisplayHour24 = SimpleDLS::GetDisplayHour24(rtc.m_Hour24, rtc.m_DayOfMonth, rtc.m_Month, rtc.m_Year);
  Config(DisplayHour24, 0, 23, '0', 3);
  Config(rtc.m_Minute, 0, 59, '0', 4);
  rtc.m_Hour24 = SimpleDLS::GetClockHour24(DisplayHour24, rtc.m_DayOfMonth, rtc.m_Month, rtc.m_Year);
  rtc.m_Second = 0;
  rtc.WriteTime();
  
  Splash(1000, false);

  rtc.ReadTime();
  rtc.ReadDate();
}

void Clock::Mode()
{
  // cycle the mode
  m_DisplayMode++;
  if (m_DisplayMode >= eModeLAST)
  {
    m_DisplayMode = 0;
  }
  rtc.WriteByte(CLOCK_MODE_INDEX, m_DisplayMode);
  
  m_Mesh.Clear();
  m_Mesh.Display(pMEW, 4, 8, 1);
  m_Mesh.Paint();
  delay(1000);
  m_Mesh.Clear();
}
