#include <avr/pgmspace.h>
#include "TriangleMesh.h"
#include "fastTFT.h"

// Triangle Mesh:
//  Start with a grid of square cells.  
//  Display characters using a simple 4x4 font by turning cells "on" or "off".
//  Then randomly shift the top left corner of each cell a little in X & Y.
//  (Distort less at the edges of the grid and where characters are displayed.)
//  Divide the cell (now a quadrilateral) into two triangles with a random diagonal.
//  When painting the cells, start with a random base colour as HSV (Hue Saturation and Value).
//  Change the hue by a random amount across and down the grid.
//  Shade the tirangular facets of the quadrilateral by altering the Value so the triangles look lit from top left.
//  Highlight the "on" cells by taking the hue the cell would have and using the complimentary hue instead.
//  
//  A redraw of the cells is either
//  1: A full redraw: a left-to-right-and-down scan of every cell, required when the grid has been randomised.  
//  The exposed black areas around the edges are dealt with separately.
//  2: A partial redraw: each cell is repainted, done when the colours are re-randomized.  
//  A Linear feedback shift register is used to visit every cell in a random order, rather than a scan.
//  The exposed black areas around the edges remaind untouched.
//  3: possibly an individual cell (and it's neighbours) are redrawn in response to a touch.
//  
//  The quadrilaterals are drawn as two triangles by evaluating Bresenham's line algorithm for each line.
//  Instead of drawing within the algorithm, arrays of left-most and right-most X coordinates, for a given Y, are updated.
//  These are then used to draw a series of horizontal lines on the lcd.
//  
//  fastTFT is designed to draw these lines as fast as possible.


#define swap(_a, _b) {long t=_a;_a=_b;_b=t;}


// simple 4x5 cell character definitions, packed into DWORDs
// 5 nibbles, each nibble is a row, most significant bit is left-most cell
const prog_uint32_t TriangleMesh::charDefs[] PROGMEM = 
{
  // '/'= thin '1', sans serifs RHS
  0x44444,
  // 0..9
  0x75557, 0x62227,0x71747,0x71717,0x55711,0x74717, 0x74757,0x71111,0x75757,0x75717,
  // ':'=':'
  0x04040,
  // ';'=<degrees>
  0x75700,
  // '<'= thin '1', centred
  0x22222
};


TriangleMesh::TriangleMesh():
  flags(0)
{
}

TriangleMesh::~TriangleMesh()
{
}

void TriangleMesh::Clear()
{
  ClearCellStates();
}

// display the given string
int TriangleMesh::Display(const char* Str, int row, int col)
{
  while (*Str && col < cellsAcross)
  {
    SetChar(GetCharDefinition(*Str), row, col);
    Str++;
  }
  return col;
}

// display the given char definitions
void TriangleMesh::Display(prog_uint32_t* Defns, int Count, int row, int col)
{
  while (col < cellsAcross && Count)
  {
    SetChar(pgm_read_dword(Defns), row, col);
    Defns++;
    Count--;
  }
}


// repaint the whole grid
#ifdef _WINDOWS
void TriangleMesh::Paint(CDC& dc)
{
  TFT_pDC = &dc;
#else
void TriangleMesh::Paint()
{
#endif
  // rasterizing of triangles (in pairs)
  char aX[maxVertexY];
  char bX[maxVertexY];
  char cX[maxVertexY];

  // init each row to N/A
  for (int i = 0; i < maxVertexY; i++)
  {
    aX[i] = bX[i] = cX[i] = vertexNULL;
  }

  if (flags & newMeshFlag)
  {
    // blank around the outside, could be smarter and less noticable, but would complicate & slow the code
    // blank upper
    fastTFT::SetXY(0, 0);
    fastTFT::BlackPixels(screenWidth*(edgeHeight + maxVertexOffset));

    // blank left
    fastTFT::SetXY(0, edgeHeight + maxVertexOffset);
    fastTFT::SetLandscapeMargin(edgeWidth + maxVertexOffset - 1);
    fastTFT::BlackPixels((edgeWidth + maxVertexOffset)*(screenHeight - 2*(edgeHeight + maxVertexOffset)));
    
    // blank right
    fastTFT::SetXY(screenWidth - edgeWidth - maxVertexOffset, edgeHeight + maxVertexOffset);
    fastTFT::SetLandscapeMargin(screenWidth - 1);
    fastTFT::BlackPixels((edgeWidth + maxVertexOffset)*(screenHeight - 2*(edgeHeight + maxVertexOffset)));

    // blank lower
    fastTFT::SetXY(0, screenHeight - edgeHeight - maxVertexOffset);
    fastTFT::BlackPixels(screenWidth*(edgeHeight + maxVertexOffset));

    // paint the cells left to right and top down
    for (int row = 0; row < cellsDown; row++)
    {
      for (int col = 0; col < cellsAcross; col++)
      {
        PaintCell(col, row, aX, bX, cX);
      }
    }
    flags &= ~newMeshFlag;
  }
  else
  {
    // recolour the existing cells, visit the cells in a pseudo-random order
    // http://en.wikipedia.org/wiki/Linear_feedback_shift_register
    byte LFSR = (byte)cellColourS;
    if (!LFSR) 
    {
      // seed must be non-0
      LFSR = 0xCA;
    }
    for (int i = 0; i < 255; i++)
    {
      // taps @ 8 6 5 4, period 255
      byte bit  = ((LFSR >> 0) ^ (LFSR >> 2) ^ (LFSR >> 3) ^ (LFSR >> 4) ) & 1;
      LFSR = (LFSR >> 1) | (bit << 7);
      if (LFSR <= cellsAcross*cellsDown)
      {
        // we see LFSR in the range 1..cellsAcross*cellsDown
        int col = (LFSR % cellsAcross);
        int row = (LFSR / cellsAcross) % cellsDown;
        PaintCell(col, row, aX, bX, cX);
      }
    }
  }
}

// virtual, returns true if the display should be dimmed (eg night-time)
bool TriangleMesh::DimDisplay()
{
  return false;
}

// paint the given cell
void TriangleMesh::PaintCell(int col, int row, char* aX, char* bX, char* cX)
{
  int x0 = col*cellWidth  - maxVertexOffset;
  int y0 = row*cellHeight - maxVertexOffset;

  int x1,y1,x2,y2,x3,y3,x4,y4;  // 1 2
                                // 4 3

  GetVertex(x0, y0, col,   row,   x1, y1);
  GetVertex(x0, y0, col+1, row,   x2, y2);
  GetVertex(x0, y0, col+1, row+1, x3, y3);
  GetVertex(x0, y0, col,   row+1, x4, y4);

  // don't overlap the cells
  x2--;
  y4--;
  x3--;
  y3--;

  byte cell = cells[col][row];
  bool on = (cell & cellStateBit) != 0;
  bool diagonal = (cell & cellDiagonalBit) != 0;

  Colour rgb1;
  Colour rgb2;
  GetCellColours(col, row, on, diagonal, rgb1, rgb2);

  // compute the rows of pixels
  if (diagonal)  // diagonal is '/'
  {
    // 1 2
    // 4
    Bresenham(x1, y1, x2, y2, aX, bX);
    Bresenham(x2, y2, x4, y4, NULL, bX);  // the diagonal can't have left-most pixels
    Bresenham(x4, y4, x1, y1, aX, NULL);  // the vertical can't have right-most pixels

    //   2
    // 4 3
    Bresenham(x2, y2, x3, y3, NULL, cX);  // the vertical can't have left-most pixels
    Bresenham(x3, y3, x4, y4, bX, cX);
    // (don't do the diagonal again)
  }
  else
  {
    // "left" triangle before "right" one; swap colour
    swap(rgb1, rgb2);

    // 1
    // 4 3
    Bresenham(x1, y1, x3, y3, NULL, bX);
    Bresenham(x3, y3, x4, y4, aX, bX);
    Bresenham(x4, y4, x1, y1, aX, NULL);

    // 1 2
    //   3
    Bresenham(x1, y1, x2, y2, bX, cX);
    Bresenham(x2, y2, x3, y3, NULL, cX);
  }

  // draw the lines
  int startX = x0 + edgeWidth;
  int startY = y0 + edgeHeight;
  fastTFT::SetColour(rgb1);
  fastTFT::SetAltColour(rgb2);

  for (int i = 0; i < maxVertexY; i++)
  {
    if (aX[i] != vertexNULL) // there's a LHS
    {
      fastTFT::SetXY(startX + aX[i], startY);
      fastTFT::ColourPixels(bX[i] - aX[i] + 1);

      if (cX[i] > bX[i])
      {
        // there's a RHS and it's more than 1 pixel
        // let the LHS have the shared pixel
        fastTFT::AltColourPixels(cX[i] - bX[i]);
      }
      // clear as we go
      aX[i] = bX[i] = cX[i] = vertexNULL;
    }
    else if (cX[i] != vertexNULL) // there's just a RHS
    {
      fastTFT::SetXY(startX + bX[i], startY);
      fastTFT::AltColourPixels(cX[i] - bX[i] + 1);
      // clear as we go
      bX[i] = cX[i] = vertexNULL;
    }

    startY++;
  }
}

#ifdef BYTE_HSV
Colour TriangleMesh::HSVtoRGB(ColourComponent h, ColourComponent s, ColourComponent v)
{
  ColourComponent region;
  unsigned short r, g, b, fpart;

  // one of 6 hue regions
  region = h / 43;
  // remainder within region, scaled 0-255
  fpart = (h - (region * 43)) * 6L;

  switch (region)
  {
    case 0:
      r = v;
      g = (v * (255 - ((s * (255 - fpart)) >> 8))) >> 8; // t
      b = (v * (255 - s)) >> 8; // p
      break;
    case 1:
      r = (v * (255 - ((s * fpart) >> 8))) >> 8; // q
      g = v;
      b = (v * (255 - s)) >> 8; // p
      break;
    case 2:
      r = (v * (255 - s)) >> 8; // p
      g = v;
      b = (v * (255 - ((s * (255 - fpart)) >> 8))) >> 8;  // t
      break;
    case 3:
      r = (v * (255 - s)) >> 8; // p
      g = (v * (255 - ((s * fpart) >> 8))) >> 8; // q
      b = v;
      break;
    case 4:
      r = (v * (255 - ((s * (255 - fpart)) >> 8))) >> 8;  // t
      g = (v * (255 - s)) >> 8; // p
      b = v;
      break;
    case 5:
      r = v;
      g = (v * (255 - s)) >> 8; // p
      b = (v * (255 - ((s * fpart) >> 8))) >> 8; // q
      break;
    default:
      return 0;
  }

  // for simplicity, just truncate
#ifdef _WINDOWS
  return RGB(r, g, b);
#else
  // "BGR" for the TFT todo
  return ((r & 0xF8) >> 3) |
         ((g & 0xFC) << 3) |
         ((b & 0xF8) << 8);
#endif
}
#else
// convert Hue, Saturation & Value to Red, Green & Blue
// based on web.mit.edu/storborg/Public/hsvtorgb.c, refactored
// * 16-bit integers rather than 8-bit
// * DON'T calculate temp vars (p, q, t), only those needed for the RGB
// * no grayscale, S != 0
// * combine scaling back with building 565
Colour TriangleMesh::HSVtoRGB(ColourComponent h, ColourComponent s, ColourComponent v)
{
  ColourComponent region;
  unsigned long r, g, b, fpart;

  // one of 6 hue regions
  region = h / 10923L;
  // remainder within region, scaled 0-65535
  fpart = (h - (region * 10923L)) * 6L;

  switch (region)
  {
    case 0:
      r = v;
      g = (unsigned long)(v * (65535L - ((s * (65535L - fpart)) >> 16))) >> 16; // t
      b = (unsigned long)(v * (65535L - s)) >> 16; // p
      break;
    case 1:
      r = (unsigned long)(v * (65535L - ((s * fpart) >> 16))) >> 16; // q
      g = v;
      b = (unsigned long)(v * (65535L - s)) >> 16; // p
      break;
    case 2:
      r = (unsigned long)(v * (65535L - s)) >> 16; // p
      g = v;
      b = (unsigned long)(v * (65535L - ((s * (65535L - fpart)) >> 16))) >> 16;  // t
      break;
    case 3:
      r = (unsigned long)(v * (65535L - s)) >> 16; // p
      g = (unsigned long)(v * (65535L - ((s * fpart) >> 16))) >> 16; // q
      b = v;
      break;
    case 4:
      r = (unsigned long)(v * (65535L - ((s * (65535L - fpart)) >> 16))) >> 16;  // t
      g = (unsigned long)(v * (65535L - s)) >> 16; // p
      b = v;
      break;
    case 5:
      r = v;
      g = (unsigned long)(v * (65535L - s)) >> 16; // p
      b = (unsigned long)(v * (65535L - ((s * fpart) >> 16))) >> 16; // q
      break;
    default:
      return 0;
  }

  // for simplicity, just truncate
#ifdef _WINDOWS
  return RGB(r >> 8, g >> 8, b >> 8);
#else
  // "BGR" for the TFT
  return ( b >> 11) |
         ((g & 0xFC00) >> 5)  |
         ( r & 0xF800);
#endif
}
#endif

// randomize the colours across the mesh
void TriangleMesh::RandomizeColour(bool smear)
{
  if (smear)
    flags |= smearColoursFlag;
  else
    flags &= ~smearColoursFlag;
  // base seed colour
#ifdef BYTE_HSV
  cellColourH = random(0x100);      // random Hue 0..255
  cellColourS = random(0x7F, 0x100);  // random Saturation 50..99, or 75..99?, 0xBFFF
  // random value 50..(100-MaxLightening)
  cellColourV = random(0x7F, 0xFF - maxDeltaV*2);
#else
  cellColourH = random(0x10000);      // random Hue 0..359
  cellColourS = random(0x7FFF, 0x10000);  // random Saturation 50..99, or 75..99?, 0xBFFF
  // random value 50..(100-MaxLightening)
  cellColourV = random(0x7FFF, 0xFFFF - maxDeltaV*2);
#endif
  if (DimDisplay())
  {
    cellColourV /= 4;
    cellColourV *= 3;
  }

  // the delta hue per column, -maxDeltaH..maxDeltaH
  cellDeltaHCol = maxDeltaH - random(2*maxDeltaH);
  // the delta hue per row, -maxDeltaH..maxDeltaH
  cellDeltaHRow = maxDeltaH - random(2*maxDeltaH);
}

// get the two colours for the triangles in the cell
void TriangleMesh::GetCellColours(int col, int row, bool on, bool diagonal, Colour& rgb1, Colour& rgb2)
{
  ColourComponent h = cellColourH;
  ColourComponent s = cellColourS;
  // shade the triangle faces
  ColourComponent V1 = cellColourV + maxDeltaV/((diagonal)?1:2);
  ColourComponent V2 = cellColourV - maxDeltaV/((diagonal)?1:2);

  if (flags & smearColoursFlag)
  {
    h += (ColourComponent)(row*cellDeltaHRow + col*cellDeltaHCol);
  }

  if (on)
  {
    // make the on cells stand out, use the complimentary colour
#ifdef BYTE_HSV
    h += 0x7F;
#else
    h += 0x7FFF;
#endif
  }

  rgb1 = HSVtoRGB(h, s, V1);
  rgb2 = HSVtoRGB(h, s, V2);
}

// virtual, returns the 3x5 font definition for Char, or 0
unsigned long TriangleMesh::GetCharDefinition(char Char)
{
  if ('/' <= Char && Char <= '<')
  {
    return 	pgm_read_dword(charDefs + Char - '/');
  }
  return 0x00000;
}

// set the cell states to display the character definition at the position on the grid
void TriangleMesh::SetChar(unsigned long charDef, int baseRow, int& baseCol)
{
  unsigned long Mask = 0x100000;
  for (int Y = 0; Y < fontHeight; Y++)
    for (int X = 0; X < fontWidth; X++)
      if (charDef & (Mask >>= 1))
        cells[baseCol + X][baseRow - fontHeight + Y + 1] |= cellStateBit;
      else
        cells[baseCol + X][baseRow - fontHeight + Y + 1] &= ~cellStateBit;

  baseCol += fontWidth;
}

// zero the states of all the cells
void TriangleMesh::ClearCellStates()
{
  for (int row = 0; row <= cellsDown; row++)
  {
    for (int col = 0; col <= cellsAcross; col++)
    {
      // DXXXSYYY, clear S
      cells[col][row] &= ~cellStateBit;
    }
  }
}

// randomize (distort) the mesh
void TriangleMesh::RandomizeMesh(bool distort)
{
  if (distort)
    flags |= distortVerticesFlag;
  else
    flags &= ~distortVerticesFlag;
  for (int row = 0; row <= cellsDown; row++)
  {
    for (int col = 0; col <= cellsAcross; col++)
    {
      RandomizeCell(cells[col][row]);
    }
  }
  flags |= newMeshFlag;
}


// distort the cell corner at the position
bool TriangleMesh::RandomizeNode(int x, int y)
{
  // only if a redraw is fast enough
  static int prevCol = -1;
  static int prevRow = -1;
  int col = x / cellWidth;
  int row = y / cellHeight;
  if (0 <= col && col < cellsAcross && 0 <= row && row < cellsDown)
  {
    if (row != prevRow || col != prevCol)
    {
      prevRow = row;
      prevCol = col;
      RandomizeCell(cells[col][row]);
      return true;
    }
  }
  return false;
}

// randomize a cell's position and diagonal direction
void TriangleMesh::RandomizeCell(byte& cell)
{
  // DXXXSYYY
  cell &= cellStateBit;  // preserve the cell state (S)
  cell |= random() & ~cellStateBit; // random Diagonal, dX & dY
}

// get the amount of distortion encoded in the nibble, up to max
int TriangleMesh::CellOffset(byte nibble, int max)
{
  int offset = (nibble & 0x07) + 1;  // 1..8
  if (offset > maxVertexOffset)
  {
    // -1..-4
    offset = maxVertexOffset - offset;
    if (offset < -max)
    {
      offset = -max;
    }
  }
  else
  {
    // +1..+4
    if (offset > +max)
    {
      offset = +max;
    }
  }
  return offset;
}

// get the (top left) corner of the given cell, normalized relative to x0, y0
void TriangleMesh::GetVertex(int x0, int y0, int col, int row, int& x, int& y)
{
  x = col*cellWidth - x0;
  y = row*cellHeight - y0;
  if (flags & distortVerticesFlag)
  {
    byte cell = cells[col][row];
    int max = maxVertexOffset;
    
    if (4 <= row && row <= 9)
    {
      // On/Off cells are distorted less
      max >>= 1;
      if (cell & cellStateBit)
      {
        // On cells are distorted less still
        max >>= 1;
      }
    }

    if (0 < col && col < cellsAcross)
    {
      x += CellOffset(cell >> 4, max);
    }
    else
    {
      // edges are special cases, shift less to stay on screen
      x += CellOffset(cell >> 4, edgeWidth - 1);
    }

    if (0 < row && row < cellsDown)
    {
      y += CellOffset(cell, max);
    }
    else
    {
      // edges are special cases, shift less to stay on screen
      y += CellOffset(cell, edgeHeight - 1);
    }
  }
}

/*
       ____
    a-/    ----b___    <- a, b, triangle 1 only
     /   1   ___-- /
  a-/ ___-b--     /-c  <- a, b, c, triangle 1 & 2
   /-___     2   /
       b----____/-c    <- b, c, triangle 2 only
*/
// "virtually" render a line, updating the arrays of minimum and maximum x coordinates, normalised by x0, y0
void TriangleMesh::Bresenham(int x0, int y0,int x1, int y1, char* min, char* max)
{
  // based on http://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
  bool steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep)
  {
    swap(x0, y0);
    swap(x1, y1);
  }
  if (x0 > x1)
  {
    swap(x0, x1);
    swap(y0, y1);
  }
  int deltax = x1 - x0;
  int deltay = abs(y1 - y0);
  int error = deltax / 2;
  int ystep;
  int y = y0;
  if (y0 < y1)
    ystep = 1;
  else 
    ystep = -1;
  for (int x = x0; x <= x1; x++)
  {
    int plotX = x;
    int plotY = y;
    if (steep)
      swap(plotX,plotY);

    // This is where we would plot(plotX,plotY) if we were drawing the line.
    // just update the row info instead
    if (min)
    {
      if (min[plotY] == vertexNULL)
      {
        // first time, init the row
        min[plotY] = plotX;
      }
      else
      {
        // update min
        if (plotX < min[plotY])
          min[plotY] = plotX;
      }
    }

    if (max && plotX > max[plotY]) // update max
    {
      // Note: no need to check against vertexNULL, it's -99
      max[plotY] = plotX;
    }

    error -= deltay;
    if (error < 0)
    {
      y += ystep;
      error += deltax;
    }
  }
}

