#ifndef _fastTFT_h
#define _fastTFT_h
#include <TFT.h>

// a fast implementation of basic TFT commands
// aimed at horizontal line fills in one of two colours, in landscape mode

// send a command, called with a manifest constant, upper byte is 0
#define FAST_TFT_COMMAND(_cmd) PORTD = 0;PORTB = RD_BIT;PORTB = RD_BIT | WR_BIT;\
                               PORTD = (byte)(_cmd << 2);PORTB = (byte)(_cmd >> 6) | RD_BIT;PORTB = (byte)(_cmd >> 6) | RD_BIT | WR_BIT;

class fastTFT
{
  public:
    // Initialise the TFT. No fill. 
    // portrait is the default orientation, landscape is portrait rotated 90deg clockwise.
    static void Init(bool portrait, int delayMS);  // TFT library uses 100ms
    
    // PORTB:
    //   00RWRCdd 
    //   --DRSS76
    // PORTD:
    //   dddddd00
    //   543210--
    // send a word of data
    static inline void SendData(word data)
    {
      byte temp1;
      byte temp2;
      // could be even faster using PINx registers?
      PORTD = temp2 = (data >> 6);
      PORTB = temp1 = (data >> 14) | RD_BIT | RS_BIT;
      PORTB = temp1 | WR_BIT;
      
      PORTD = (data << 2);
      PORTB = temp1 = (temp2 & 0x03) | RD_BIT | RS_BIT;
      PORTB = temp1 | WR_BIT;
    }

    // set the pixel cursor
    static inline void SetXYPortrait(word X, word Y)
    {
      // in portrait mode, set the DRAM address
      FAST_TFT_COMMAND(0x20);
      SendData(X);
      FAST_TFT_COMMAND(0x21);
      SendData(Y);
      FAST_TFT_COMMAND(0x22);  // ready to write colour data
    }

    static inline void SetXYLandscape(word X, word Y)
    {
      // in landscape, set the corner of the window
      FAST_TFT_COMMAND(0x53);
      SendData(MAX_Y - X);
      FAST_TFT_COMMAND(0x50);
      SendData(Y);
      FAST_TFT_COMMAND(0x22);  // ready to write colour data
    }

    static inline void SetLandscapeMargin(word X)
    {
      // in landscape, set the right edge of the window
      // MOVES THE CURSOR to the top-left corner of the window too
      FAST_TFT_COMMAND(0x52);
      SendData(MAX_Y - X);
      FAST_TFT_COMMAND(0x22);  // ready to write colour data
    }
    
    // set the current colour
    static inline void SetColour(word colour)
    {
      // pre-compute the port values
      // this will write lower 6 bits of colour's high byte:
      portDHi = (colour >> 6);
      // this will write the upper two bits of colour's high byte, AND set RD+RS, AND clear CS and WR
      portBHi = (colour >> 14) | RD_BIT | RS_BIT;
      // as above, but with with WR set:
      portBHiWrite = portBHi | WR_BIT;

      // as above, but the low byte      
      portDLo = (colour << 2);
      portBLo = (portDHi & 0x03) | RD_BIT | RS_BIT;
      portBLoWrite = portBLo | WR_BIT;
    }

    // set the alternate colour
    static inline void SetAltColour(word colour)
    {
      // pre-compute the port values
      // this will write lower 6 bits of colour's high byte:
      altPortDHi = (colour >> 6);
      // this will write the upper two bits of colour's high byte, AND set RD+RS, AND clear CS and WR
      altPortBHi = (colour >> 14) | RD_BIT | RS_BIT;
      // as above, but with with WR set:
      altPortBHiWrite = altPortBHi | WR_BIT;

      // as above, but the low byte      
      altPortDLo = (colour << 2);
      altPortBLo = (altPortDHi & 0x03) | RD_BIT | RS_BIT;
      altPortBLoWrite = altPortBLo | WR_BIT;
    }
    
    // sets the next count pixels to the current colour
    static inline void ColourPixels(long count)
    {
      while (count--)
      {
        PORTD = portDHi;
        PORTB = portBHi;
        PORTB = portBHiWrite;

        PORTD = portDLo;
        PORTB = portBLo;
        PORTB = portBLoWrite;
      }
    };

    // sets the next count pixels to the alternate colour
    static inline void AltColourPixels(long count)
    {
      while (count--)
      {
        PORTD = altPortDHi;
        PORTB = altPortBHi;
        PORTB = altPortBHiWrite;

        PORTD = altPortDLo;
        PORTB = altPortBLo;
        PORTB = altPortBLoWrite;
      }
    };
    
    // sets the next count pixels to black
    static inline void BlackPixels(long count)
    {
      PORTD = 0;
      while (count--)
      {
        // special, slightly faster case when the data is all 0's
        PORTB = RD_BIT | RS_BIT;
        PORTB = RD_BIT | RS_BIT | WR_BIT;
        PORTB = RD_BIT | RS_BIT;
        PORTB = RD_BIT | RS_BIT | WR_BIT;
      }
    }

    // sets the next count pixels to white
    static inline void WhitePixels(long count)
    {
      PORTD = 0xFC;
      while (count--)
      {
        // special, slightly faster case when the data is all FF's
        PORTB = 0x03 | RD_BIT | RS_BIT;
        PORTB = 0x03 | RD_BIT | RS_BIT | WR_BIT;
        PORTB = 0x03 | RD_BIT | RS_BIT;
        PORTB = 0x03 | RD_BIT | RS_BIT | WR_BIT;
      }
    }
    
  private:
    static byte portDHi, portBHi, portBHiWrite, portDLo, portBLo, portBLoWrite;
    static byte altPortDHi, altPortBHi, altPortBHiWrite, altPortDLo, altPortBLo, altPortBLoWrite;
    static prog_uint16_t initData[] PROGMEM;
};

#endif
